from traceback import format_exc

from django.conf import settings

from graphite.logger import log
from graphite.readers.utils import BaseReader
from graphite.util import unpickle, msgpack, BufferedHTTPReader

import time


class MeasuredReader(object):
  def __init__(self, reader):
    self.reader = reader
    self.bytes_read = 0

  def read(self, amt=None):
    b = b''
    try:
      if amt:
        b = self.reader.read(amt)
      else:
        b = self.reader.read()
      return b
    finally:
      self.bytes_read += len(b)


class RemoteReader(BaseReader):
  __slots__ = (
    'finder',
    'metric_path',
    'intervals',
    'bulk_query',
  )

  def __init__(self, finder, node_info, bulk_query=None):
    self.finder = finder
    self.metric_path = node_info.get('path') or node_info.get('metric_path')
    self.intervals = node_info.get('intervals', [])
    self.bulk_query = set(bulk_query) if bulk_query else (
      [self.metric_path] if self.metric_path else []
    )

  def __repr__(self):
    return '<RemoteReader[%x]: %s %s>' % (id(self), self.finder.host, ','.join(self.bulk_query))

  def get_intervals(self):
    return self.intervals

  def fetch(self, startTime, endTime, now=None, requestContext=None):
    for series in self.fetch_multi(startTime, endTime, now, requestContext):
      if series['name'] == self.metric_path:
        return (series['time_info'], series['values'])

  def fetch_multi(self, startTime, endTime, now=None, requestContext=None):
    if not self.bulk_query:
      return []

    query_params = [
      ('format', self.finder.params.get('format', 'pickle')),
      ('local', self.finder.params.get('local', '1')),
      ('noCache', '1'),
      ('from', int(startTime)),
      ('until', int(endTime))
    ]

    for target in self.bulk_query:
      query_params.append(('target', target))

    if now is not None:
      query_params.append(('now', int(now)))

    headers = requestContext.get('forwardHeaders') if requestContext else None

    retries = 1  # start counting at one to make log output and settings more readable
    while True:
      try:
        result = self.finder.request(
          '/render/',
          fields=query_params,
          headers=headers,
          timeout=settings.FETCH_TIMEOUT,
        )
        break
      except Exception:
        if retries >= settings.MAX_FETCH_RETRIES:
          log.exception("Failed after %s attempts! Root cause:\n%s" %
                        (settings.MAX_FETCH_RETRIES, format_exc()))
          raise
        else:
          log.exception("Got an exception when fetching data! Try: %i of %i. Root cause:\n%s" %
                        (retries, settings.MAX_FETCH_RETRIES, format_exc()))
        retries += 1

    data = self.deserialize(result)

    try:
      return [
        {
          'pathExpression': series.get('pathExpression', series['name']),
          'name': series['name'],
          'time_info': (series['start'], series['end'], series['step']),
          'values': series['values'],
        }
        for series in data
      ]
    except Exception as err:
      self.finder.fail()
      log.exception(
        "RemoteReader[%s] Invalid render response from %s: %s" %
        (self.finder.host, result.url_full, repr(err)))
      raise Exception("Invalid render response from %s: %s" % (result.url_full, repr(err)))

  def deserialize(self, result):
    """
    Based on configuration, either stream-deserialize a response in settings.REMOTE_BUFFER_SIZE chunks,
    or read the entire payload and use inline deserialization.
    :param result: an http response object
    :return: deserialized response payload from cluster server
    """
    start = time.time()
    try:
      should_buffer = settings.REMOTE_BUFFER_SIZE > 0
      measured_reader = MeasuredReader(BufferedHTTPReader(result, settings.REMOTE_BUFFER_SIZE))

      if should_buffer:
        log.debug("Using streaming deserializer.")
        reader = BufferedHTTPReader(measured_reader, settings.REMOTE_BUFFER_SIZE)
        deserialized = self._deserialize_stream(reader, result.getheader('content-type'))
        return deserialized
      else:
        log.debug("Using inline deserializer for small payload")
        deserialized = self._deserialize_buffer(measured_reader.read(), result.getheader('content-type'))
        return deserialized
    except Exception as err:
      self.finder.fail()
      log.exception(
        "RemoteReader[%s] Error decoding render response from %s: %s" %
        (self.finder.host, result.url_full, err))
      raise Exception("Error decoding render response from %s: %s" % (result.url_full, err))
    finally:
      log.debug("Processed %d bytes in %f seconds." % (measured_reader.bytes_read, time.time() - start))
      result.release_conn()

  @staticmethod
  def _deserialize_buffer(byte_buffer, content_type):
    if content_type == 'application/x-msgpack':
      data = msgpack.unpackb(byte_buffer, encoding='utf-8')
    else:
      data = unpickle.loads(byte_buffer)

    return data

  @staticmethod
  def _deserialize_stream(stream, content_type):
    if content_type == 'application/x-msgpack':
      data = msgpack.load(stream, encoding='utf-8')
    else:
      data = unpickle.load(stream)

    return data



from django.db import models


class Series(models.Model):
  hash = models.CharField(max_length=64, unique=True)
  path = models.TextField()
  __str__ = lambda self: "Series [%s]" % self.path


class Tag(models.Model):
  tag = models.CharField(max_length=191, unique=True)
  __str__ = lambda self: "Tag [%s]" % self.tag


class TagValue(models.Model):
  value = models.CharField(max_length=191, unique=True)
  __str__ = lambda self: "TagValue [%s]" % self.value


class SeriesTag(models.Model):
  series = models.ForeignKey(Series, on_delete=models.CASCADE)
  tag = models.ForeignKey(Tag, on_delete=models.CASCADE)
  value = models.ForeignKey(TagValue, on_delete=models.CASCADE)

  class Meta:
    unique_together = ("series", "tag")

  __str__ = lambda self: "SeriesTag [%s %s %s]" % (self.series, self.tag, self.value)



from graphite.util import jsonResponse, HttpResponse, HttpError
from graphite.functions import SeriesFunctions, SeriesFunction, PieFunctions, PieFunction, functionInfo


def jsonEncoder(obj):
  if hasattr(obj, 'toJSON'):
    return obj.toJSON()
  return obj.__dict__


@jsonResponse(default=jsonEncoder)
def functionList(request, queryParams):
  if request.method != 'GET':
    return HttpResponse(status=405)

  if queryParams.get('type') == 'pie':
    funcs = PieFunctions()
  else:
    funcs = SeriesFunctions()

  grouped = queryParams.get('grouped', '').lower() in ['1', 'true']
  group = queryParams.get('group')
  result = {}

  for (name, func) in funcs.items():
    info = functionInfo(name, func)
    if group is not None and group != info['group']:
      continue

    if grouped:
      if info['group'] not in result:
        result[info['group']] = {}
      result[info['group']][name] = info
    else:
      result[name] = info

  return result


@jsonResponse(default=jsonEncoder)
def functionDetails(request, queryParams, name):
  if request.method != 'GET':
    return HttpResponse(status=405)

  try:
    if queryParams.get('type') == 'pie':
      func = PieFunction(name)
    else:
      func = SeriesFunction(name)
  except KeyError:
    raise HttpError('Function not found: %s' % name, status=404)

  return functionInfo(name, func)





class Node(object):
  __slots__ = ('name', 'path', 'local', 'is_leaf')

  def __init__(self, path):
    self.path = path
    self.name = path.split('.')[-1]
    self.local = True
    self.is_leaf = False

  def __repr__(self):
    return '<%s[%x]: %s>' % (self.__class__.__name__, id(self), self.path)


class BranchNode(Node):
  pass


class LeafNode(Node):
  __slots__ = ('reader', )

  def __init__(self, path, reader):
    Node.__init__(self, path)
    self.reader = reader
    self.is_leaf = True

  def fetch(self, startTime, endTime, now=None, requestContext=None):
    try:
      result = self.reader.fetch(startTime, endTime, now, requestContext)
    except TypeError:
      # Support for legacy 3rd party, readers.
      result = self.reader.fetch(startTime, endTime)

    return result

  @property
  def intervals(self):
    return self.reader.get_intervals()

  def __repr__(self):
    return '<LeafNode[%x]: %s (%s)>' % (id(self), self.path, self.reader)



from django.shortcuts import render_to_response
from graphite import settings


def index(request):
  context = {
    'version' : settings.WEBAPP_VERSION,
  }
  return render_to_response('version.html', context)



from __future__ import absolute_import

import re
import bisect
import sys

from graphite.tags.base import BaseTagDB, TaggedSeries


class RedisTagDB(BaseTagDB):
  """
  Stores tag information in a Redis database.

  Keys used are:

  .. code-block:: none

    series                    # Set of all paths
    series:<path>:tags        # Hash of all tag:value pairs for path
    tags                      # Set of all tags
    tags:<tag>:series         # Set of paths with entry for tag
    tags:<tag>:values         # Set of values for tag
    tags:<tag>:values:<value> # Set of paths matching tag/value

  """
  def __init__(self, settings, *args, **kwargs):
    super(RedisTagDB, self).__init__(settings, *args, **kwargs)

    from redis import Redis

    self.r = Redis(
      host=settings.TAGDB_REDIS_HOST,
      port=settings.TAGDB_REDIS_PORT,
      db=settings.TAGDB_REDIS_DB,
      decode_responses=(sys.version_info[0] >= 3),
    )

  def _find_series(self, tags, requestContext=None):
    selector = None
    selector_cnt = None
    filters = []

    # loop through tagspecs, look for best spec to use as selector
    for tagspec in tags:
      (tag, operator, spec) = self.parse_tagspec(tagspec)

      if operator == '=':
        matches_empty = spec == ''
        if not matches_empty:
          cnt = self.r.scard('tags:' + tag + ':values:' + spec)
          if not selector or selector[1] != '=' or selector_cnt > cnt:
            if selector:
              filters.append(selector)
            selector = (tag, operator, spec)
            selector_cnt = cnt
            continue
        filters.append((tag, operator, spec))

      elif operator == '=~':
        pattern = re.compile(spec)
        matches_empty = bool(pattern.match(''))
        if not matches_empty and (not selector or selector[1] != '='):
          cnt = self.r.scard('tags:' + tag + ':values')
          if not selector or selector_cnt > cnt:
            if selector:
              filters.append(selector)
            selector = (tag, operator, pattern)
            selector_cnt = cnt
            continue
        filters.append((tag, operator, pattern))

      elif operator == '!=':
        matches_empty = spec != ''
        if not matches_empty and (not selector or selector[1] != '='):
          cnt = self.r.scard('tags:' + tag + ':values')
          if not selector or selector_cnt > cnt:
            if selector:
              filters.append(selector)
            selector = (tag, operator, spec)
            selector_cnt = cnt
            continue
        filters.append((tag, operator, spec))

      elif operator == '!=~':
        pattern = re.compile(spec)
        matches_empty = not pattern.match('')
        if not matches_empty and (not selector or selector[1] != '='):
          cnt = self.r.scard('tags:' + tag + ':values')
          if not selector or selector_cnt > cnt:
            if selector:
              filters.append(selector)
            selector = (tag, operator, pattern)
            selector_cnt = cnt
            continue
        filters.append((tag, operator, pattern))

      else:
        raise ValueError("Invalid operator %s" % operator)

    if not selector:
      raise ValueError("At least one tagspec must not match the empty string")

    # get initial list of series
    (tag, operator, spec) = selector

    # find list of values that match the tagspec
    values = None
    if operator == '=':
      values = [spec]
    elif operator == '=~':
      # see if we can identify a literal prefix to filter by in redis
      match = None
      m = re.match('([a-z0-9]+)([^*?|][^|]*)?$', spec.pattern)
      if m:
        match = m.group(1) + '*'
      values = [value for value in self.r.sscan_iter('tags:' + tag + ':values', match=match) if spec.match(value) is not None]
    elif operator == '!=':
      values = [value for value in self.r.sscan_iter('tags:' + tag + ':values') if value != spec]
    elif operator == '!=~':
      values = [value for value in self.r.sscan_iter('tags:' + tag + ':values') if spec.match(value) is None]

    # if this query matched no values, just short-circuit since the result of the final intersect will be empty
    if not values:
      return []

    results = []

    # apply filters
    operators = ['=','!=','=~','!=~']
    filters.sort(key=lambda a: operators.index(a[1]))

    for series in self.r.sunion(*['tags:' + tag + ':values:' + value for value in values]):
      parsed = self.parse(series)
      matched = True

      for (tag, operator, spec) in filters:
        value = parsed.tags.get(tag, '')
        if (
          (operator == '=' and value != spec) or
          (operator == '=~' and spec.match(value) is None) or
          (operator == '!=' and value == spec) or
          (operator == '!=~' and spec.match(value) is not None)
        ):
          matched = False
          break

      if matched:
        bisect.insort_left(results, series)

    return results

  def get_series(self, path, requestContext=None):
    tags = {}

    tags = self.r.hgetall('series:' + path + ':tags')
    if not tags:
      return None

    return TaggedSeries(tags['name'], tags)

  def list_tags(self, tagFilter=None, limit=None, requestContext=None):
    result = []

    if tagFilter:
      tagFilter = re.compile(tagFilter)

    for tag in self.r.sscan_iter('tags'):
      if tagFilter and tagFilter.match(tag) is None:
        continue
      if len(result) == 0 or tag >= result[-1]:
        if limit and len(result) >= limit:
          continue
        result.append(tag)
      else:
        bisect.insort_left(result, tag)
      if limit and len(result) > limit:
        del result[-1]

    return [
      {'tag': tag}
      for tag in result
    ]

  def get_tag(self, tag, valueFilter=None, limit=None, requestContext=None):
    if not self.r.sismember('tags', tag):
      return None

    return {
      'tag': tag,
      'values': self.list_values(
        tag,
        valueFilter=valueFilter,
        limit=limit,
        requestContext=requestContext
      ),
    }

  def list_values(self, tag, valueFilter=None, limit=None, requestContext=None):
    result = []

    if valueFilter:
      valueFilter = re.compile(valueFilter)

    for value in self.r.sscan_iter('tags:' + tag + ':values'):
      if valueFilter and valueFilter.match(value) is None:
        continue
      if len(result) == 0 or value >= result[-1]:
        if limit and len(result) >= limit:
          continue
        result.append(value)
      else:
        bisect.insort_left(result, value)
      if limit and len(result) > limit:
        del result[-1]

    return [
      {'value': value, 'count': self.r.scard('tags:' + tag + ':values:' + value)}
      for value in result
    ]

  def tag_series(self, series, requestContext=None):
    # extract tags and normalize path
    parsed = self.parse(series)

    path = parsed.path

    with self.r.pipeline() as pipe:
      pipe.sadd('series', path)

      for tag, value in parsed.tags.items():
        pipe.hset('series:' + path + ':tags', tag, value)

        pipe.sadd('tags', tag)
        pipe.sadd('tags:' + tag + ':series', path)
        pipe.sadd('tags:' + tag + ':values', value)
        pipe.sadd('tags:' + tag + ':values:' + value, path)

      pipe.execute()

    return path

  def del_series(self, series, requestContext=None):
    # extract tags and normalize path
    parsed = self.parse(series)

    path = parsed.path

    with self.r.pipeline() as pipe:
      pipe.srem('series', path)

      pipe.delete('series:' + path + ':tags')

      for tag, value in parsed.tags.items():
        pipe.srem('tags:' + tag + ':series', path)
        pipe.srem('tags:' + tag + ':values:' + value, path)

      pipe.execute()

    return True



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

from django.db import models
from django.contrib.auth import models as auth_models


class Profile(models.Model):
  user = models.OneToOneField(auth_models.User, on_delete=models.CASCADE)
  history = models.TextField(default="")
  advancedUI = models.BooleanField(default=False)
  __str__ = lambda self: "Profile for %s" % self.user


class Variable(models.Model):
  profile = models.ForeignKey(Profile, on_delete=models.CASCADE)
  name = models.CharField(max_length=64)
  value = models.CharField(max_length=64)


class View(models.Model):
  profile = models.ForeignKey(Profile, on_delete=models.CASCADE)
  name = models.CharField(max_length=64)


class Window(models.Model):
  view = models.ForeignKey(View, on_delete=models.CASCADE)
  name = models.CharField(max_length=64)
  top = models.IntegerField()
  left = models.IntegerField()
  width = models.IntegerField()
  height = models.IntegerField()
  url = models.TextField()
  interval = models.IntegerField(null=True)


class MyGraph(models.Model):
  profile = models.ForeignKey(Profile, on_delete=models.CASCADE)
  name = models.CharField(max_length=64)
  url = models.TextField()



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

from django.contrib.auth import authenticate, login, logout
try:
    from django.urls import reverse
except ImportError:  # Django < 1.10
    from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.shortcuts import render_to_response
from graphite.user_util import getProfile, isAuthenticated


def loginView(request):
  username = request.POST.get('username')
  password = request.POST.get('password')
  if request.method == 'GET':
    nextPage = request.GET.get('nextPage', reverse('browser'))
  else:
    nextPage = request.POST.get('nextPage', reverse('browser'))
  if username and password:
    user = authenticate(username=username,password=password)
    if user is None:
      return render_to_response("login.html",{'authenticationFailed' : True, 'nextPage' : nextPage})
    elif not user.is_active:
      return render_to_response("login.html",{'accountDisabled' : True, 'nextPage' : nextPage})
    else:
      login(request,user)
      return HttpResponseRedirect(nextPage)
  else:
    return render_to_response("login.html",{'nextPage' : nextPage})


def logoutView(request):
  nextPage = request.GET.get('nextPage', reverse('browser'))
  logout(request)
  return HttpResponseRedirect(nextPage)


def editProfile(request):
  if not isAuthenticated(request.user):
    return HttpResponseRedirect(reverse('browser'))
  context = { 'profile' : getProfile(request) }
  return render_to_response("editProfile.html",context)


def updateProfile(request):
  profile = getProfile(request,allowDefault=False)
  if profile:
    profile.advancedUI = request.POST.get('advancedUI','off') == 'on'
    profile.save()
  nextPage = request.POST.get('nextPage', reverse('browser'))
  return HttpResponseRedirect(nextPage)



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import os, logging
from logging.handlers import TimedRotatingFileHandler as Rotater
try:
    from logging import NullHandler
except ImportError as ie:  # py2.6
    from logging import Handler

    class NullHandler(Handler):

        def emit(self, record):
            pass
try:
    from logging import FileHandler, StreamHandler
except ImportError as ie:  # py2.6
    from logging.handlers import FileHandler, StreamHandler
from django.conf import settings


class GraphiteLogger:
  def __init__(self):
    self.infoLogger = self._config_logger(
        settings.LOG_FILE_INFO,
        'info',
        True,
        level = logging.DEBUG if settings.DEBUG else logging.INFO,
    )
    self.exceptionLogger = self._config_logger(
        settings.LOG_FILE_EXCEPTION,
        'exception',
        True,
    )
    self.cacheLogger = self._config_logger(
        settings.LOG_FILE_CACHE,
        'cache',
        settings.LOG_CACHE_PERFORMANCE,
        level = logging.DEBUG if settings.DEBUG else logging.INFO,
    )
    self.renderingLogger = self._config_logger(
        settings.LOG_FILE_RENDERING,
        'rendering',
        settings.LOG_RENDERING_PERFORMANCE,
        level = logging.DEBUG if settings.DEBUG else logging.INFO,
    )

  @staticmethod
  def _config_logger(log_file_name, name, activate,
                     level=None, when='midnight',
                     backupCount=settings.LOG_ROTATION_COUNT):
    logger = logging.getLogger(name)
    if level is not None:
        logger.setLevel(level)
    if activate:  # if want to log this one
        if log_file_name == '-':
            formatter = logging.Formatter(
                fmt='[%(asctime)s.%(msecs)03d] %(name)s %(levelname)s %(message)s',
                datefmt='%d/%b/%Y %H:%M:%S')
            handler = StreamHandler()
        else:
            formatter = logging.Formatter(
                fmt='%(asctime)s.%(msecs)03d :: %(message)s',
                datefmt='%Y-%m-%d,%H:%M:%S')
            log_file = os.path.join(settings.LOG_DIR, log_file_name)
            if settings.LOG_ROTATION:  # if we want to rotate logs
                handler = Rotater(log_file, when=when, backupCount=backupCount)
            else:  # let someone else, e.g. logrotate, rotate the logs
                handler = FileHandler(log_file)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    else:
        logger.addHandler(NullHandler())
    return logger

  def info(self,msg,*args,**kwargs):
    return self.infoLogger.info(msg,*args,**kwargs)

  def debug(self,msg,*args,**kwargs):
    return self.infoLogger.debug(msg,*args,**kwargs)

  def warning(self,msg,*args,**kwargs):
    return self.infoLogger.warn(msg,*args,**kwargs)

  def exception(self,msg="Exception Caught",**kwargs):
    return self.exceptionLogger.exception(msg,**kwargs)

  def cache(self,msg,*args,**kwargs):
    return self.cacheLogger.info(msg,*args,**kwargs)

  def rendering(self,msg,*args,**kwargs):
    return self.renderingLogger.info(msg,*args,**kwargs)


log = GraphiteLogger() # import-shared logger instance



import time
import socket
import struct
import random

from django.conf import settings

from graphite.render.hashing import ConsistentHashRing
from graphite.logger import log
from graphite.util import load_module, unpickle
from graphite.singleton import ThreadSafeSingleton


try:
  import six.moves.cPickle as pickle
except ImportError:
  import pickle


def load_keyfunc():
  if settings.CARBONLINK_HASHING_KEYFUNC:
    module_path, func_name = settings.CARBONLINK_HASHING_KEYFUNC.rsplit(':', 1)
    log.cache("Using keyfunc %s found in %s" % (str(func_name), str(module_path)))
    return load_module(module_path, member=func_name)
  else:
    return lambda x: x


class CarbonLinkRequestError(Exception):
  pass


class CarbonLinkPool(object):
  def __init__(self, hosts, timeout):
    self.hosts = [ (server, instance) for (server, port, instance) in hosts ]
    self.ports = dict(
      ((server, instance), port) for (server, port, instance) in hosts )
    self.timeout = float(timeout)
    servers = set([server for (server, port, instance) in hosts])
    if len(servers) < settings.REPLICATION_FACTOR:
      raise Exception("REPLICATION_FACTOR=%d cannot exceed servers=%d" % (
        settings.REPLICATION_FACTOR, len(servers)))

    self.hash_ring = ConsistentHashRing(
      self.hosts, hash_type=settings.CARBONLINK_HASHING_TYPE)
    self.keyfunc = load_keyfunc()
    self.connections = {}
    self.last_failure = {}
    # Create a connection pool for each host
    for host in self.hosts:
      self.connections[host] = set()

  def select_host(self, metric):
    "Returns the carbon host that has data for the given metric"
    key = self.keyfunc(metric)
    nodes = []
    servers = set()
    for node in self.hash_ring.get_nodes(key):
      (server, instance) = node
      if server in servers:
        continue
      servers.add(server)
      nodes.append(node)
      if len(servers) >= settings.REPLICATION_FACTOR:
        break

    available = [ n for n in nodes if self.is_available(n) ]
    return random.choice(available or nodes)

  def is_available(self, host):
    now = time.time()
    last_fail = self.last_failure.get(host, 0)
    return (now - last_fail) < settings.CARBONLINK_RETRY_DELAY

  def get_connection(self, host):
    # First try to take one out of the pool for this host
    (server, instance) = host
    port = self.ports[host]
    connectionPool = self.connections[host]
    try:
      return connectionPool.pop()
    except KeyError:
      pass #nothing left in the pool, gotta make a new connection

    log.cache("CarbonLink creating a new socket for %s" % str(host))
    connection = socket.socket()
    connection.settimeout(self.timeout)
    try:
      connection.connect((server, port))
    except socket.error:
      self.last_failure[host] = time.time()
      raise
    else:
      connection.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
      return connection

  def query(self, metric):
    request = dict(type='cache-query', metric=metric)
    results = self.send_request(request)
    log.cache("CarbonLink cache-query request for %s returned %d datapoints" % (
      metric, len(results['datapoints'])))
    return results['datapoints']

  def get_metadata(self, metric, key):
    request = dict(type='get-metadata', metric=metric, key=key)
    results = self.send_request(request)
    log.cache("CarbonLink get-metadata request received for %s:%s" % (metric, key))
    return results['value']

  def set_metadata(self, metric, key, value):
    request = dict(type='set-metadata', metric=metric, key=key, value=value)
    results = self.send_request(request)
    log.cache("CarbonLink set-metadata request received for %s:%s" % (metric, key))
    return results

  def send_request(self, request):
    metric = request['metric']
    serialized_request = pickle.dumps(request, protocol=-1)
    len_prefix = struct.pack("!L", len(serialized_request))
    request_packet = len_prefix + serialized_request
    result = {}
    result.setdefault('datapoints', [])

    if metric.startswith(settings.CARBON_METRIC_PREFIX):
      return self.send_request_to_all(request)

    if not self.hosts:
      log.cache("CarbonLink is not connected to any host. Returning empty nodes list")
      return result

    host = self.select_host(metric)
    conn = self.get_connection(host)
    log.cache("CarbonLink sending request for %s to %s" % (metric, str(host)))
    try:
      conn.sendall(request_packet)
      result = self.recv_response(conn)
    except Exception as e:
      self.last_failure[host] = time.time()
      log.cache("Exception getting data from cache %s: %s" % (str(host), e))
    else:
      self.connections[host].add(conn)
      if 'error' in result:
        log.cache("Error getting data from cache: %s" % result['error'])
        raise CarbonLinkRequestError(result['error'])
      log.cache("CarbonLink finished receiving %s from %s" % (str(metric), str(host)))
    return result

  def send_request_to_all(self, request):
    metric = request['metric']
    serialized_request = pickle.dumps(request, protocol=-1)
    len_prefix = struct.pack("!L", len(serialized_request))
    request_packet = len_prefix + serialized_request
    results = {}
    results.setdefault('datapoints', {})

    for host in self.hosts:
      conn = self.get_connection(host)
      log.cache("CarbonLink sending request for %s to %s" % (metric, str(host)))
      try:
        conn.sendall(request_packet)
        result = self.recv_response(conn)
      except Exception as e:
        self.last_failure[host] = time.time()
        log.cache("Exception getting data from cache %s: %s" % (str(host), e))
      else:
        self.connections[host].add(conn)
        if 'error' in result:
          log.cache("Error getting data from cache %s: %s" % (str(host), result['error']))
        else:
          if len(result['datapoints']) > 1:
              results['datapoints'].update(result['datapoints'])
      log.cache("CarbonLink finished receiving %s from %s" % (str(metric), str(host)))
    return results

  def recv_response(self, conn):
    len_prefix = self.recv_exactly(conn, 4)
    body_size = struct.unpack("!L", len_prefix)[0]
    body = self.recv_exactly(conn, body_size)
    return unpickle.loads(body)

  @staticmethod
  def recv_exactly(conn, num_bytes):
    buf = b''
    while len(buf) < num_bytes:
      data = conn.recv(num_bytes - len(buf))
      if not data:
        raise Exception("Connection lost")
      buf += data

    return buf


@ThreadSafeSingleton
class GlobalCarbonLinkPool(CarbonLinkPool):
  def __init__(self):
    hosts = []
    for host in settings.CARBONLINK_HOSTS:
      parts = host.split(':')
      server = parts[0]
      port = int(parts[1])
      if len(parts) > 2:
        instance = parts[2]
      else:
        instance = None
      hosts.append((server, int(port), instance))
    timeout = settings.CARBONLINK_TIMEOUT
    CarbonLinkPool.__init__(self, hosts, timeout)


def CarbonLink():
  """Handy accessor for the global singleton."""
  return GlobalCarbonLinkPool.instance()



"""Utility functions for tag databases."""
import re

from hashlib import sha256


class TaggedSeries(object):
  @classmethod
  def parse(cls, path):
    # if path is in openmetrics format: metric{tag="value",...}
    if path[-2:] == '"}' and '{' in path:
      return cls.parse_openmetrics(path)

    # path is a carbon path with optional tags: metric;tag=value;...
    return cls.parse_carbon(path)

  @classmethod
  def parse_openmetrics(cls, path):
    """parse a path in openmetrics format: metric{tag="value",...}

    https://github.com/RichiH/OpenMetrics
    """
    (metric, rawtags) = path[0:-1].split('{', 2)
    if not metric:
      raise Exception('Cannot parse path %s, no metric found' % path)

    tags = {}

    while len(rawtags) > 0:
      m = re.match(r'([^=]+)="((?:[\\]["\\]|[^"\\])+)"(:?,|$)', rawtags)
      if not m:
        raise Exception('Cannot parse path %s, invalid segment %s' % (path, rawtags))

      tags[m.group(1)] = m.group(2).replace(r'\"', '"').replace(r'\\', '\\')
      rawtags = rawtags[len(m.group(0)):]

    tags['name'] = metric
    return cls(metric, tags)

  @classmethod
  def parse_carbon(cls, path):
    """parse a carbon path with optional tags: metric;tag=value;..."""
    segments = path.split(';')

    metric = segments[0]
    if not metric:
      raise Exception('Cannot parse path %s, no metric found' % path)

    tags = {}

    for segment in segments[1:]:
      tag = segment.split('=', 1)
      if len(tag) != 2 or not tag[0]:
        raise Exception('Cannot parse path %s, invalid segment %s' % (path, segment))

      tags[tag[0]] = tag[1]

    tags['name'] = metric
    return cls(metric, tags)

  @staticmethod
  def format(tags):
    return tags.get('name', '') + ''.join(sorted([
      ';%s=%s' % (tag, value)
      for tag, value in tags.items()
      if tag != 'name'
    ]))

  @staticmethod
  def encode(metric, sep='.', hash_only=False):
    """
    Helper function to encode tagged series for storage in whisper etc

    When tagged series are detected, they are stored in a separate hierarchy of folders under a
    top-level _tagged folder, where subfolders are created by using the first 3 hex digits of the
    sha256 hash of the tagged metric path (4096 possible folders), and second-level subfolders are
    based on the following 3 hex digits (another 4096 possible folders) for a total of 4096^2
    possible subfolders. The metric files themselves are created with any . in the metric path
    replaced with -, to avoid any issues where metrics, tags or values containing a '.' would end
    up creating further subfolders. This helper is used by both whisper and ceres, but by design
    each carbon database and graphite-web finder is responsible for handling its own encoding so
    that different backends can create their own schemes if desired.

    The hash_only parameter can be set to True to use the hash as the filename instead of a
    human-readable name.  This avoids issues with filename length restrictions, at the expense of
    being unable to decode the filename and determine the original metric name.

    A concrete example:

    .. code-block:: none

      some.metric;tag1=value2;tag2=value.2

      with sha256 hash starting effaae would be stored in:

      _tagged/eff/aae/some-metric;tag1=value2;tag2=value-2.wsp (whisper)
      _tagged/eff/aae/some-metric;tag1=value2;tag2=value-2 (ceres)

    """
    if ';' in metric:
      metric_hash = sha256(metric.encode('utf8')).hexdigest()
      return sep.join([
        '_tagged',
        metric_hash[0:3],
        metric_hash[3:6],
        metric_hash if hash_only else metric.replace('.', '_DOT_')
      ])

    # metric isn't tagged, just replace dots with the separator and trim any leading separator
    return metric.replace('.', sep).lstrip(sep)

  @staticmethod
  def decode(path, sep='.'):
    """
    Helper function to decode tagged series from storage in whisper etc
    """
    if path.startswith('_tagged'):
      return path.split(sep, 3)[-1].replace('_DOT_', '.')

    # metric isn't tagged, just replace the separator with dots
    return path.replace(sep, '.')

  def __init__(self, metric, tags, series_id=None):
    self.metric = metric
    self.tags = tags
    self.id = series_id

  @property
  def path(self):
    return self.__class__.format(self.tags)



import re

from django.db import connection
from hashlib import sha256

from graphite.tags.base import BaseTagDB, TaggedSeries


class LocalDatabaseTagDB(BaseTagDB):
  def find_series_query(self, tags):
    # sql will select series that match all tag expressions that don't match empty tags
    sql = 'SELECT s.path'
    sql += ' FROM tags_series AS s'
    params = []

    where = []
    whereparams = []

    all_match_empty = True

    # expressions that do match empty tags will be used to filter the result
    filters = []

    i = 0
    for tagspec in tags:
      (tag, operator, spec) = self.parse_tagspec(tagspec)

      i += 1
      s = str(i)

      if operator == '=':
        matches_empty = spec == ''

        if not matches_empty:
          where.append('v' + s + '.value=%s')
          whereparams.append(spec)

      elif operator == '=~':
        # make sure regex is anchored
        if not spec.startswith('^'):
          spec = '^(' + spec + ')'

        matches_empty = bool(re.match(spec, ''))

        if not matches_empty:
          where.append('v' + s + '.value ' + self._regexp_operator(connection) + ' %s')
          whereparams.append(spec)

      elif operator == '!=':
        matches_empty = spec != ''

        if not matches_empty:
          where.append('v' + s + '.value<>%s')
          whereparams.append(spec)

      elif operator == '!=~':
        # make sure regex is anchored
        if not spec.startswith('^'):
          spec = '^(' + spec + ')'

        matches_empty = not re.match(spec, '')

        if not matches_empty:
          where.append('v' + s + '.value ' + self._regexp_not_operator(connection) + ' %s')
          whereparams.append(spec)

      else:
        raise ValueError("Invalid operator %s" % operator)

      if matches_empty:
        filters.append((tag, operator, spec))
      else:
        sql += ' JOIN tags_tag AS t' + s + ' ON t' + s + '.tag=%s'
        params.append(tag)
        sql += ' JOIN tags_seriestag AS st' + s + ' ON st' + s + '.series_id=s.id AND st' + s + '.tag_id=t' + s + '.id'
        sql += ' JOIN tags_tagvalue AS v' + s + ' ON v' + s + '.id=st' + s + '.value_id'

      all_match_empty = all_match_empty and matches_empty

    if all_match_empty:
      raise ValueError("At least one tagspec must not match the empty string")

    if where:
      sql += ' WHERE ' + ' AND '.join(where)
      params.extend(whereparams)

    sql += ' ORDER BY s.path'

    return sql, params, filters

  def _find_series(self, tags, requestContext=None):
    sql, params, filters = self.find_series_query(tags)

    def matches_filters(path):
      if not filters:
        return True

      parsed = self.parse(path)

      for (tag, operator, spec) in filters:
        value = parsed.tags.get(tag, '')
        if (
          (operator == '=' and value != spec) or
          (operator == '=~' and re.match(spec, value) is None) or
          (operator == '!=' and value == spec) or
          (operator == '!=~' and re.match(spec, value) is not None)
        ):
          return False

      return True

    with connection.cursor() as cursor:
      cursor.execute(sql, params)

      return [row[0] for row in cursor if matches_filters(row[0])]

  def get_series(self, path, requestContext=None):
    with connection.cursor() as cursor:
      sql = 'SELECT s.id, t.tag, v.value'
      sql += ' FROM tags_series AS s'
      sql += ' JOIN tags_seriestag AS st ON st.series_id=s.id'
      sql += ' JOIN tags_tag AS t ON t.id=st.tag_id'
      sql += ' JOIN tags_tagvalue AS v ON v.id=st.value_id'
      sql += ' WHERE s.path=%s'
      params = [path]
      cursor.execute(sql, params)

      series_id = None

      tags = {tag: value for (series_id, tag, value) in cursor}

      if not tags:
        return None

      return TaggedSeries(tags['name'], tags, series_id=series_id)

  def list_tags(self, tagFilter=None, limit=None, requestContext=None):
    with connection.cursor() as cursor:
      sql = 'SELECT t.id, t.tag'
      sql += ' FROM tags_tag AS t'
      params = []

      if tagFilter:
        # make sure regex is anchored
        if not tagFilter.startswith('^'):
          tagFilter = '^(' + tagFilter + ')'
        sql += ' WHERE t.tag ' + self._regexp_operator(connection) + ' %s'
        params.append(tagFilter)

      sql += ' ORDER BY t.tag'

      if limit:
        sql += ' LIMIT %s'
        params.append(int(limit))

      cursor.execute(sql, params)

      return [{'id': tag_id, 'tag': tag} for (tag_id, tag) in cursor]

  def get_tag(self, tag, valueFilter=None, limit=None, requestContext=None):
    with connection.cursor() as cursor:
      sql = 'SELECT t.id, t.tag'
      sql += ' FROM tags_tag AS t'
      sql += ' WHERE t.tag=%s'
      params = [tag]
      cursor.execute(sql, params)

      row = cursor.fetchone()

    if not row:
      return None

    (tag_id, tag) = row

    return {
      'id': tag_id,
      'tag': tag,
      'values': self.list_values(
        tag,
        valueFilter=valueFilter,
        limit=limit,
        requestContext=requestContext
      ),
    }

  def list_values(self, tag, valueFilter=None, limit=None, requestContext=None):
    with connection.cursor() as cursor:
      sql = 'SELECT v.id, v.value, COUNT(st.id)'
      sql += ' FROM tags_tagvalue AS v'
      sql += ' JOIN tags_seriestag AS st ON st.value_id=v.id'
      sql += ' JOIN tags_tag AS t ON t.id=st.tag_id'
      sql += ' WHERE t.tag=%s'
      params = [tag]

      if valueFilter:
        # make sure regex is anchored
        if not valueFilter.startswith('^'):
          valueFilter = '^(' + valueFilter + ')'
        sql += ' AND v.value ' + self._regexp_operator(connection) + ' %s'
        params.append(valueFilter)

      sql += ' GROUP BY v.id, v.value'
      sql += ' ORDER BY v.value'

      if limit:
        sql += ' LIMIT %s'
        params.append(int(limit))

      cursor.execute(sql, params)

      return [{'id': value_id, 'value': value, 'count': count} for (value_id, value, count) in cursor]

  @staticmethod
  def _insert_ignore(table, cols, data):
    sql = table + ' (' + ','.join(cols) + ') VALUES ' + ', '.join(['(' + ', '.join(['%s'] * len(cols)) + ')'] * len(data))
    params = []
    for row in data:
      params.extend(row)

    if connection.vendor == 'mysql':
      sql = 'INSERT IGNORE INTO ' + sql
    elif connection.vendor == 'sqlite':
      sql = 'INSERT OR IGNORE INTO ' + sql
    elif connection.vendor == 'postgresql':
      sql = 'INSERT INTO ' + sql + ' ON CONFLICT DO NOTHING'  # nosec
    else:
      raise Exception('Unsupported database vendor ' + connection.vendor)

    with connection.cursor() as cursor:
      cursor.execute(sql, params)

  @staticmethod
  def _regexp_operator(connection):
    if connection.vendor == 'mysql':
      return 'REGEXP'
    if connection.vendor == 'sqlite':
      # django provides an implementation of REGEXP for sqlite
      return 'REGEXP'
    if connection.vendor == 'postgresql':
      return '~*'
    raise Exception('Database vendor ' + connection.vendor + ' does not support regular expressions')

  @staticmethod
  def _regexp_not_operator(connection):
    if connection.vendor == 'mysql':
      return 'NOT REGEXP'
    if connection.vendor == 'sqlite':
      # django provides an implementation of REGEXP for sqlite
      return 'NOT REGEXP'
    if connection.vendor == 'postgresql':
      return '!~*'
    raise Exception('Database vendor ' + connection.vendor + ' does not support regular expressions')

  def tag_series(self, series, requestContext=None):
    # extract tags and normalize path
    parsed = self.parse(series)

    path = parsed.path

    # check if path is already tagged
    curr = self.get_series(path)
    if curr and parsed.tags == curr.tags:
      return path

    with connection.cursor() as cursor:
      # tags
      self._insert_ignore('tags_tag', ['tag'], [[tag] for tag in parsed.tags.keys()])

      sql = 'SELECT id, tag FROM tags_tag WHERE tag IN (' + ', '.join(['%s'] * len(parsed.tags)) + ')'  # nosec
      params = list(parsed.tags.keys())
      cursor.execute(sql, params)
      tag_ids = {tag: tag_id for (tag_id, tag) in cursor}

      # tag values
      self._insert_ignore('tags_tagvalue', ['value'], [[value] for value in parsed.tags.values()])

      sql = 'SELECT id, value FROM tags_tagvalue WHERE value IN (' + ', '.join(['%s'] * len(parsed.tags)) + ')'  # nosec
      params = list(parsed.tags.values())
      cursor.execute(sql, params)
      value_ids = {value: value_id for (value_id, value) in cursor}

      # series
      if curr:
        series_id = curr.id
      else:
        # hash column is used to support a unique index in mysql since path can be longer than 191 characters
        path_hash = sha256(path.encode('utf8')).hexdigest()
        self._insert_ignore('tags_series', ['hash', 'path'], [[path_hash, path]])

        sql = 'SELECT id FROM tags_series WHERE path=%s'
        params = [path]
        cursor.execute(sql, params)
        series_id = cursor.fetchone()[0]

      # series tags
      self._insert_ignore(
        'tags_seriestag',
        ['series_id', 'tag_id', 'value_id'],
        [[series_id, tag_ids[tag], value_ids[value]] for tag, value in parsed.tags.items()]
      )

    return path

  def del_series(self, series, requestContext=None):
    # extract tags and normalize path
    parsed = self.parse(series)

    path = parsed.path

    with connection.cursor() as cursor:
      sql = 'SELECT id'
      sql += ' FROM tags_series'
      sql += ' WHERE path=%s'
      params = [path]
      cursor.execute(sql, params)

      row = cursor.fetchone()
      if not row:
        return True

      (series_id, ) = row

      sql = 'DELETE FROM tags_seriestag WHERE series_id=%s'
      params = [series_id]
      cursor.execute(sql, params)

      sql = 'DELETE FROM tags_series WHERE id=%s'
      params = [series_id]
      cursor.execute(sql, params)

    return True



from django.db import models
from graphite.account.models import Profile
from graphite.util import json
import six


class Dashboard(models.Model):
  name = models.CharField(primary_key=True, max_length=128)
  owners = models.ManyToManyField(Profile, related_name='dashboards')
  state = models.TextField()
  __str__ = lambda self: "Dashboard [%s]" % self.name


class Template(models.Model):

  class Admin: pass
  name = models.CharField(primary_key=True, max_length=128)
  owners = models.ManyToManyField(Profile, related_name='templates')
  state = models.TextField()
  __str__ = lambda self: "Template [%s]" % self.name

  def loadState(self, val):
    return self.state.replace('__VALUE__', val)

  def setState(self, state, key):
    #XXX Might not need this
    def replace_string(s):
      if isinstance(s, six.text_type):
        s = s.replace(key, '__VALUE__')
      return s

    def update_graph(graph):
      graph_opts = graph[1]
      graph_opts['target'] = [replace_string(s) for s in graph_opts['target']]
      return [replace_string(graph[0]),
              graph_opts,
              replace_string(graph[2])]

    # Parse JSON here and replace first five elements of target with __VALUE__
    parsed_state = json.loads(state)
    for i, graph in enumerate(parsed_state['graphs']):
      parsed_state['graphs'][i] = update_graph(graph)
    self.state = json.dumps(parsed_state)



#!/usr/bin/env python

import sys

if sys.version_info <= (2,7):
  # SystemExit defaults to returning 1 when printing a string to stderr
  raise SystemExit("You are using python %s, but version 2.7 or greater is "
                   "required" % sys.version_info)

required = 0
optional = 0


# Test for whisper
try:
  import whisper
except ImportError:
  # No? test for ceres
  try:
    import ceres
    # We imported ceres, but not whisper so it's an optional dependency
    sys.stderr.write("[OPTIONAL] Unable to import the 'whisper' module. Without it the webapp will be unable to read .wsp files\n")
    optional += 1
  except ImportError:
    sys.stderr.write("[REQUIRED] Unable to import the 'whisper' or 'ceres' modules, please download this package from the Graphite project page and install it.\n")
    required += 1


# Test for cairocffi or pycairo
try:
  import cairocffi as cairo
except ImportError:
  sys.stderr.write("[REQUIRED] Unable to import the 'cairocffi' module, attempting to fall back to pycairo\n")
  try:
    import cairo
  except ImportError:
    sys.stderr.write("[REQUIRED] Unable to import the 'cairo' module, do you have pycairo installed for python %s?\n" % sys.version_info.major)
    cairo = None
    required += 1


# Test that pycairo has the PNG backend
try:
  if cairo:
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
    del surface
except Exception:
  sys.stderr.write("[REQUIRED] Failed to create an ImageSurface with cairo, you probably need to recompile cairo with PNG support\n")
  required += 1


# Test that cairo can find fonts
try:
  if cairo:
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
    context = cairo.Context(surface)
    context.font_extents()
    del surface, context
except Exception:
  sys.stderr.write("[REQUIRED] Failed to create text with cairo, this probably means cairo cant find any fonts. Install some system fonts and try again\n")


# Test for django
try:
  import django
except ImportError:
  sys.stderr.write("[REQUIRED] Unable to import the 'django' module, do you have Django installed for python %s?\n" % sys.version_info.major)
  django = None
  required += 1


# Test for pytz
try:
  import pytz
except ImportError:
  sys.stderr.write("[REQUIRED] Unable to import the 'pytz' module, do you have pytz module installed for python %s?\n" % sys.version_info.major)
  required += 1


# Test for pyparsing
try:
  import pyparsing
except ImportError:
  sys.stderr.write("[REQUIRED] Unable to import the 'pyparsing' module, do you have pyparsing module installed for python %s?\n" % sys.version_info.major)
  required += 1


# Test for django-tagging
try:
  import tagging
except ImportError:
  sys.stderr.write("[REQUIRED] Unable to import the 'tagging' module, do you have django-tagging installed for python %s?\n" % sys.version_info.major)
  required += 1


if django and django.VERSION[:2] < (1,8):
  sys.stderr.write("[REQUIRED] You have django version %s installed, but version 1.8 or greater is required\n" % django.get_version())
  required += 1


# Test for python-memcached
try:
  import memcache
except ImportError:
  sys.stderr.write("[OPTIONAL] Unable to import the 'memcache' module, do you have python-memcached installed for python %s? This feature is not required but greatly improves performance.\n" % sys.version_info.major)
  optional += 1


# Test for python-ldap
try:
  import ldap
except ImportError:
  sys.stderr.write("[OPTIONAL] Unable to import the 'ldap' module, do you have python-ldap installed for python %s? Without python-ldap, you will not be able to use LDAP authentication in the graphite webapp.\n" % sys.version_info.major)
  optional += 1


# Test for txamqp
try:
  import txamqp
except ImportError:
  sys.stderr.write("[OPTIONAL] Unable to import the 'txamqp' module, this is required if you want to use AMQP as an input to Carbon. Note that txamqp requires python 2.5 or greater.\n")
  optional += 1


# Test for python-rrdtool
try:
  import rrdtool
except ImportError:
  sys.stderr.write("[OPTIONAL] Unable to import the 'python-rrdtool' module, this is required for reading RRD.\n")
  optional += 1


# Test for whitenoise
try:
    import whitenoise
except ImportError:
    sys.stderr.write("[OPTIONAL] Unable to import the 'whitenoise' module. This is useful for serving static files.\n")
    optional += 1


# Test for pyhash
try:
    import pyhash
except ImportError:
    sys.stderr.write("[OPTIONAL] Unable to import the 'pyhash' module. This is useful for fnv1_ch hashing support.\n")
    optional += 1


if optional:
  sys.stderr.write("%d optional dependencies not met. Please consider the optional items before proceeding.\n" % optional)
else:
  print("All optional dependencies are met.")

if required:
  sys.stderr.write("%d necessary dependencies not met. Graphite will not function until these dependencies are fulfilled.\n" % required)
  sys.exit(1)
else:
  print("All necessary dependencies are met.")



import inspect

from importlib import import_module
from os import listdir
from os.path import dirname, join, splitext

from django.conf import settings

from graphite.functions.params import Param, ParamTypes  # noqa
from graphite.logger import log

customDir = join(dirname(__file__), 'custom')
customModPrefix = 'graphite.functions.custom.'

_SeriesFunctions = {}
_PieFunctions = {}


def loadFunctions(force=False):
  if _SeriesFunctions and not force:
    return

  from graphite.render import functions

  _SeriesFunctions.clear()
  _SeriesFunctions.update(functions.SeriesFunctions)

  _PieFunctions.clear()
  _PieFunctions.update(functions.PieFunctions)

  custom_modules = []
  for filename in listdir(customDir):
    module_name, extension = splitext(filename)
    if extension != '.py' or module_name == '__init__':
      continue
    custom_modules.append(customModPrefix + module_name)

  for module_name in custom_modules + settings.FUNCTION_PLUGINS:
    try:
      module = import_module(module_name)
    except Exception as e:
      log.warning('Error loading function plugin %s: %s' % (module_name, e))
      continue

    for func_name, func in getattr(module, 'SeriesFunctions', {}).items():
      try:
        addFunction(_SeriesFunctions, func, func_name)
      except Exception as e:
        log.warning('Error loading function plugin %s: %s' % (module_name, e))

    for func_name, func in getattr(module, 'PieFunctions', {}).items():
      try:
        addFunction(_PieFunctions, func, func_name)
      except Exception as e:
        log.warning('Error loading function plugin %s: %s' % (module_name, e))


def addFunction(dest, func, func_name):
  if not hasattr(func, 'group'):
    func.group = 'Ungrouped'

  if not hasattr(func, 'params'):
    raise Exception('No params defined for %s' % func_name)

  for param in func.params:
    if not isinstance(param, Param):
      raise Exception('Invalid param specified for %s' % func_name)
    dest[func_name] = func


def SeriesFunctions():
  loadFunctions()
  return _SeriesFunctions


def SeriesFunction(name):
  loadFunctions()
  try:
    return _SeriesFunctions[name]
  except KeyError:
    raise KeyError('Function "%s" not found' % name)


def PieFunctions():
  loadFunctions()
  return _PieFunctions


def PieFunction(name):
  loadFunctions()
  try:
    return _PieFunctions[name]
  except KeyError:
    raise KeyError('Function "%s" not found' % name)


def functionInfo(name, func):
  argspec = inspect.getargspec(func)
  return {
    'name': name,
    'function': name + inspect.formatargspec(argspec[0][1:], argspec[1], argspec[2], argspec[3]),
    'description': inspect.getdoc(func),
    'module': inspect.getmodule(func).__name__,
    'group': getattr(func, 'group', 'Ungrouped'),
    'params': getattr(func, 'params', None),
  }



import re
import errno

from os.path import getmtime
from six.moves.urllib.parse import urlencode
from six.moves.configparser import ConfigParser
from django.shortcuts import render_to_response
from django.http import QueryDict
from django.conf import settings
from django.contrib.auth import login, authenticate, logout
from django.contrib.staticfiles import finders
from django.utils.safestring import mark_safe
from graphite.compat import HttpResponse
from graphite.dashboard.models import Dashboard, Template
from graphite.dashboard.send_graph import send_graph_email
from graphite.render.views import renderView
from graphite.util import json
from graphite.user_util import isAuthenticated

fieldRegex = re.compile(r'<([^>]+)>')
defaultScheme = {
  'name' : 'Everything',
  'pattern' : '<category>',
  'fields' : [ dict(name='category', label='Category') ],
}
defaultUIConfig = {
  'default_graph_width'  : 400,
  'default_graph_height' : 250,
  'refresh_interval'     :  60,
  'autocomplete_delay'   : 375,
  'merge_hover_delay'    : 700,
  'theme'                : 'default',
}
defaultKeyboardShortcuts = {
  'toggle_toolbar' : 'ctrl-z',
  'toggle_metrics_panel' : 'ctrl-space',
  'erase_all_graphs' : 'alt-x',
  'save_dashboard' : 'alt-s',
  'completer_add_metrics' : 'alt-enter',
  'completer_del_metrics' : 'alt-backspace',
  'give_completer_focus' : 'shift-space',
}

ALL_PERMISSIONS = ['change', 'delete']


class DashboardConfig:
  def __init__(self):
    self.last_read = 0
    self.schemes = [defaultScheme]
    self.ui_config = defaultUIConfig.copy()

  def check(self):
    if getmtime(settings.DASHBOARD_CONF) > self.last_read:
      self.load()

  def load(self):
    schemes = [defaultScheme]
    parser = ConfigParser()
    parser.read(settings.DASHBOARD_CONF)

    for option, default_value in defaultUIConfig.items():
      if parser.has_option('ui', option):
        try:
          self.ui_config[option] = parser.getint('ui', option)
        except ValueError:
          self.ui_config[option] = parser.get('ui', option)
      else:
        self.ui_config[option] = default_value

    if parser.has_option('ui', 'automatic_variants'):
      self.ui_config['automatic_variants']   = parser.getboolean('ui', 'automatic_variants')
    else:
      self.ui_config['automatic_variants'] = True

    self.ui_config['keyboard_shortcuts'] = defaultKeyboardShortcuts.copy()
    if parser.has_section('keyboard-shortcuts'):
      self.ui_config['keyboard_shortcuts'].update( parser.items('keyboard-shortcuts') )

    for section in parser.sections():
      if section in ('ui', 'keyboard-shortcuts'):
        continue

      scheme = parser.get(section, 'scheme')
      fields = []

      for match in fieldRegex.finditer(scheme):
        field = match.group(1)
        if parser.has_option(section, '%s.label' % field):
          label = parser.get(section, '%s.label' % field)
        else:
          label = field

        fields.append({
          'name' : field,
          'label' : label
        })

      schemes.append({
        'name' : section,
        'pattern' : scheme,
        'fields' : fields,
      })

    self.schemes = schemes


config = DashboardConfig()


def dashboard(request, name=None):
  dashboard_conf_missing = False

  try:
    config.check()
  except OSError as e:
    if e.errno == errno.ENOENT:
      dashboard_conf_missing = True
    else:
      raise

  initialError = None
  debug = request.GET.get('debug', False)
  theme = request.GET.get('theme', config.ui_config['theme'])
  css_file = finders.find('css/dashboard-%s.css' % theme)
  if css_file is None:
    initialError = "Invalid theme '%s'" % theme
    theme = config.ui_config['theme']

  context = {
    'schemes_json': mark_safe(json.dumps(config.schemes)),
    'ui_config_json': mark_safe(json.dumps(config.ui_config)),
    'jsdebug': debug or settings.JAVASCRIPT_DEBUG,
    'debug': debug,
    'theme': theme,
    'initialError': initialError,
    'querystring': mark_safe(json.dumps(dict(request.GET.items()))),
    'dashboard_conf_missing': dashboard_conf_missing,
    'userName': '',
    'permissions': mark_safe(json.dumps(getPermissions(request.user))),
    'permissionsUnauthenticated': mark_safe(json.dumps(getPermissions(None)))
  }
  user = request.user
  if user:
      context['userName'] = user.username

  if name is not None:
    try:
      dashboard = Dashboard.objects.get(name=name)
    except Dashboard.DoesNotExist:
      context['initialError'] = "Dashboard '%s' does not exist." % name
    else:
      context['initialState'] = dashboard.state

  return render_to_response("dashboard.html", context)


def template(request, name, val):
  template_conf_missing = False

  try:
    config.check()
  except OSError as e:
    if e.errno == errno.ENOENT:
      template_conf_missing = True
    else:
      raise

  initialError = None
  debug = request.GET.get('debug', False)
  theme = request.GET.get('theme', config.ui_config['theme'])
  css_file = finders.find('css/dashboard-%s.css' % theme)
  if css_file is None:
    initialError = "Invalid theme '%s'" % theme
    theme = config.ui_config['theme']

  context = {
    'schemes_json' : json.dumps(config.schemes),
    'ui_config_json' : json.dumps(config.ui_config),
    'jsdebug' : debug or settings.JAVASCRIPT_DEBUG,
    'debug' : debug,
    'theme' : theme,
    'initialError' : initialError,
    'querystring' : json.dumps( dict( request.GET.items() ) ),
    'template_conf_missing' : template_conf_missing,
    'userName': '',
    'permissions': json.dumps(getPermissions(request.user)),
    'permissionsUnauthenticated': json.dumps(getPermissions(None))
  }

  user = request.user
  if user:
      context['userName'] = user.username

  try:
    template = Template.objects.get(name=name)
  except Template.DoesNotExist:
    context['initialError'] = "Template '%s' does not exist." % name
  else:
    state = json.loads(template.loadState(val))
    state['name'] = '%s/%s' % (name, val)
    context['initialState'] = json.dumps(state)
  return render_to_response("dashboard.html", context)


def getPermissions(user):
  """Return [change, delete] based on authorisation model and user privileges/groups"""
  if user and not isAuthenticated(user):
    user = None
  if not settings.DASHBOARD_REQUIRE_AUTHENTICATION:
    return ALL_PERMISSIONS      # don't require login
  if not user:
      return []
  # from here on, we have a user
  permissions = ALL_PERMISSIONS
  if settings.DASHBOARD_REQUIRE_PERMISSIONS:
    permissions = [permission for permission in ALL_PERMISSIONS if user.has_perm('dashboard.%s_dashboard' % permission)]
  editGroup = settings.DASHBOARD_REQUIRE_EDIT_GROUP
  if editGroup and len(user.groups.filter(name = editGroup)) == 0:
    permissions = []
  return permissions


def save(request, name):
  if 'change' not in getPermissions(request.user):
    return json_response( dict(error="Must be logged in with appropriate permissions to save") )
  # Deserialize and reserialize as a validation step
  state = str( json.dumps( json.loads( request.POST['state'] ) ) )

  try:
    dashboard = Dashboard.objects.get(name=name)
  except Dashboard.DoesNotExist:
    dashboard = Dashboard.objects.create(name=name, state=state)
  else:
    dashboard.state = state
    dashboard.save();

  return json_response( dict(success=True) )


def save_template(request, name, key):
  if 'change' not in getPermissions(request.user):
    return json_response( dict(error="Must be logged in with appropriate permissions to save the template") )
  # Deserialize and reserialize as a validation step
  state = str( json.dumps( json.loads( request.POST['state'] ) ) )

  try:
    template = Template.objects.get(name=name)
  except Template.DoesNotExist:
    template = Template.objects.create(name=name)
    template.setState(state, key)
    template.save()
  else:
    template.setState(state, key)
    template.save();

  return json_response( dict(success=True) )


def load(request, name):
  try:
    dashboard = Dashboard.objects.get(name=name)
  except Dashboard.DoesNotExist:
    return json_response( dict(error="Dashboard '%s' does not exist. " % name) )

  return json_response( dict(state=json.loads(dashboard.state)) )


def load_template(request, name, val):
  try:
    template = Template.objects.get(name=name)
  except Template.DoesNotExist:
    return json_response( dict(error="Template '%s' does not exist. " % name) )

  state = json.loads(template.loadState(val))
  state['name'] = '%s/%s' % (name, val)
  return json_response( dict(state=state) )


def delete(request, name):
  if 'delete' not in getPermissions(request.user):
    return json_response( dict(error="Must be logged in with appropriate permissions to delete") )

  try:
    dashboard = Dashboard.objects.get(name=name)
  except Dashboard.DoesNotExist:
    return json_response( dict(error="Dashboard '%s' does not exist. " % name) )
  else:
    dashboard.delete()
    return json_response( dict(success=True) )


def delete_template(request, name):
  if 'delete' not in getPermissions(request.user):
    return json_response( dict(error="Must be logged in with appropriate permissions to delete the template") )

  try:
    template = Template.objects.get(name=name)
  except Template.DoesNotExist:
    return json_response( dict(error="Template '%s' does not exist. " % name) )
  else:
    template.delete()
    return json_response( dict(success=True) )


def find(request):
  queryParams = request.GET.copy()
  queryParams.update(request.POST)

  query = queryParams.get('query', False)
  query_terms = set( query.lower().split() )
  results = []

  # Find all dashboard names that contain each of our query terms as a substring
  for dashboard_name in Dashboard.objects.order_by('name').values_list('name', flat=True):
    name = dashboard_name.lower()

    if name.startswith('temporary-'):
      continue

    found = True # blank queries return everything
    for term in query_terms:
      if term in name:
        found = True
      else:
        found = False
        break

    if found:
      results.append( dict(name=dashboard_name) )

  return json_response( dict(dashboards=results) )


def find_template(request):
  queryParams = request.GET.copy()
  queryParams.update(request.POST)

  query = queryParams.get('query', False)
  query_terms = set( query.lower().split() )
  results = []

  # Find all dashboard names that contain each of our query terms as a substring
  for template in Template.objects.all():
    name = template.name.lower()

    found = True # blank queries return everything
    for term in query_terms:
      if term in name:
        found = True
      else:
        found = False
        break

    if found:
      results.append( dict(name=template.name) )

  return json_response( dict(templates=results) )


def help(request):
  context = {}
  return render_to_response("dashboardHelp.html", context)


def email(request):
    sender = request.POST['sender']
    recipients = request.POST['recipients'].split()
    subject = request.POST['subject']
    message = request.POST['message']

    # these need to be passed to the render function in an HTTP request.
    graph_params = json.loads(request.POST['graph_params'], parse_int=str)
    target = QueryDict(urlencode({'target': graph_params.pop('target')}))
    graph_params = QueryDict(urlencode(graph_params))

    new_post = request.POST.copy()
    new_post.update(graph_params)
    new_post.update(target)
    request.POST = new_post

    resp = renderView(request)
    img = resp.content

    if img:
        attachments = [('graph.png', img, 'image/png')]
        send_graph_email(subject, sender, recipients, attachments, message)

    return json_response(dict(success=True))


def create_temporary(request):
  state = str( json.dumps( json.loads( request.POST['state'] ) ) )
  i = 0
  while True:
    name = "temporary-%d" % i
    try:
      Dashboard.objects.get(name=name)
    except Dashboard.DoesNotExist:
      dashboard = Dashboard.objects.create(name=name, state=state)
      break
    else:
      i += 1

  return json_response( dict(name=dashboard.name) )


def json_response(obj):
  return HttpResponse(content_type='application/json', content=json.dumps(obj))


def user_login(request):
  response = dict(errors={}, text={}, success=False, permissions=[])
  user = authenticate(username=request.POST['username'],
                      password=request.POST['password'])
  if user is not None:
    if user.is_active:
      login(request, user)
      response['success'] = True
      response['permissions'].extend(getPermissions(user))
    else:
      response['errors']['reason'] = 'Account disabled.'
  else:
    response['errors']['reason'] = 'Username and/or password invalid.'
  return json_response(response)


def user_logout(request):
  response = dict(errors={}, text={}, success=True)
  logout(request)
  return json_response(response)



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import pytz
from datetime import datetime, timedelta, datetime as datetimetype
from django.conf import settings

months = ['jan','feb','mar','apr','may','jun','jul','aug','sep','oct','nov','dec']
weekdays = ['sun','mon','tue','wed','thu','fri','sat']
SECONDS_STRING = 'seconds'
MINUTES_STRING = 'minutes'
HOURS_STRING = 'hours'
DAYS_STRING = 'days'
WEEKS_STRING = 'weeks'
MONTHS_STRING = 'months'
YEARS_STRING = 'years'


def parseATTime(s, tzinfo=None, now=None):
  if tzinfo is None:
    tzinfo = pytz.timezone(settings.TIME_ZONE)
  if isinstance(s, datetimetype):
    if s.tzinfo:
      return s.astimezone(tzinfo)
    return tzinfo.localize(s)

  s = s.strip().lower().replace('_','').replace(',','').replace(' ','')
  if s.isdigit():
    if len(s) == 8 and int(s[:4]) > 1900 and int(s[4:6]) < 13 and int(s[6:]) < 32:
      pass #Fall back because its not a timestamp, its YYYYMMDD form
    else:
      return datetime.fromtimestamp(int(s),tzinfo)
  if '+' in s:
    ref,offset = s.split('+',1)
    offset = '+' + offset
  elif '-' in s:
    ref,offset = s.split('-',1)
    offset = '-' + offset
  else:
    ref,offset = s,''

  return tzinfo.normalize(parseTimeReference(ref, tzinfo, now) + parseTimeOffset(offset))


def parseTimeReference(ref, tzinfo=None, now=None):
  if tzinfo is None:
    tzinfo = pytz.timezone(settings.TIME_ZONE)
  if isinstance(ref, datetimetype):
    if ref.tzinfo:
      return ref.astimezone(tzinfo)
    return tzinfo.localize(ref)

  if now is None:
    now = datetime.now(tzinfo)
  else:
    now = parseATTime(now, tzinfo)

  if not ref or ref == 'now': return now

  rawRef = ref

  # Time-of-day reference
  i = ref.find(':')
  hour,minute = 0,0
  if 0 < i < 3:
    hour = int( ref[:i] )
    minute = int( ref[i+1:i+3] )
    ref = ref[i+3:]
    if ref[:2] == 'am':
      ref = ref[2:]
    elif ref[:2] == 'pm':
      hour = (hour + 12) % 24
      ref = ref[2:]

  # Xam or XXam
  i = ref.find('am')
  if 0 < i < 3:
    hour = int( ref[:i] )
    ref = ref[i+2:]

  # Xpm or XXpm
  i = ref.find('pm')
  if 0 < i < 3:
    hour = (int( ref[:i] ) + 12) % 24
    ref = ref[i+2:]

  if ref.startswith('noon'):
    hour,minute = 12,0
    ref = ref[4:]
  elif ref.startswith('midnight'):
    hour,minute = 0,0
    ref = ref[8:]
  elif ref.startswith('teatime'):
    hour,minute = 16,0
    ref = ref[7:]

  refDate = now.replace(hour=hour,minute=minute,second=0,microsecond=0,tzinfo=None)

  # Day reference
  if ref in ('yesterday','today','tomorrow'): # yesterday, today, tomorrow
    if ref == 'yesterday':
      refDate -= timedelta(days=1)
    elif ref == 'tomorrow':
      refDate += timedelta(days=1)

  elif ref.count('/') == 2: # MM/DD/YY[YY]
    m,d,y = map(int,ref.split('/'))
    if y < 1900: y += 1900
    if y < 1970: y += 100
    refDate = datetime(year=y,month=m,day=d,hour=hour,minute=minute)

  elif len(ref) == 8 and ref.isdigit(): # YYYYMMDD
    refDate = datetime(year=int(ref[:4]),month=int(ref[4:6]),day=int(ref[6:8]),hour=hour,minute=minute)

  elif ref[:3] in months: # MonthName DayOfMonth
    d = None
    if ref[-2:].isdigit():
      d = int(ref[-2:])
    elif ref[-1:].isdigit():
      d = int(ref[-1:])
    else:
      raise Exception("Day of month required after month name")
    refDate = datetime(year=refDate.year,month=months.index(ref[:3]) + 1,day=d,hour=hour,minute=minute)

  elif ref[:3] in weekdays: # DayOfWeek (Monday, etc)
    todayDayName = refDate.strftime("%a").lower()[:3]
    today = weekdays.index( todayDayName )
    twoWeeks = weekdays * 2
    dayOffset = today - twoWeeks.index(ref[:3])
    if dayOffset < 0: dayOffset += 7
    refDate -= timedelta(days=dayOffset)

  elif ref:
    raise ValueError("Unknown day reference: %s" % rawRef)

  return tzinfo.localize(refDate)


def parseTimeOffset(offset):
  if not offset:
    return timedelta()

  t = timedelta()

  if offset[0].isdigit():
    sign = 1
  else:
    try:
      sign = { '+' : 1, '-' : -1 }[offset[0]]
    except KeyError:
      raise KeyError('Invalid offset: %s' % offset)
    offset = offset[1:]

  while offset:
    i = 1
    while offset[:i].isdigit() and i <= len(offset): i += 1
    num = int(offset[:i-1])
    offset = offset[i-1:]
    i = 1
    while offset[:i].isalpha() and i <= len(offset): i += 1
    unit = offset[:i-1]
    offset = offset[i-1:]
    unitString = getUnitString(unit)
    if unitString == MONTHS_STRING:
      unitString = DAYS_STRING
      num = num * 30
    if unitString == YEARS_STRING:
      unitString = DAYS_STRING
      num = num * 365
    t += timedelta(**{ unitString : sign * num})

  return t


def getUnitString(s):
  if s.startswith('s'): return SECONDS_STRING
  if s.startswith('min'): return MINUTES_STRING
  if s.startswith('h'): return HOURS_STRING
  if s.startswith('d'): return DAYS_STRING
  if s.startswith('w'): return WEEKS_STRING
  if s.startswith('mon'): return MONTHS_STRING
  if s.startswith('y'): return YEARS_STRING
  raise Exception("Invalid offset unit '%s'" % s)



import re
import six

from graphite.errors import NormalizeEmptyResultError
from graphite.functions import SeriesFunction
from graphite.render.grammar import grammar
from graphite.render.datalib import fetchData, TimeSeries, prefetchData


def evaluateTarget(requestContext, targets):
  if not isinstance(targets, list):
    targets = [targets]

  pathExpressions = extractPathExpressions(requestContext, targets)
  prefetchData(requestContext, pathExpressions)

  seriesList = []

  for target in targets:
    if not target:
      continue

    if isinstance(target, six.string_types):
      if not target.strip():
        continue
      target = grammar.parseString(target)

    result = evaluateTokens(requestContext, target)

    # we have to return a list of TimeSeries objects
    if isinstance(result, TimeSeries):
      seriesList.append(result)
    elif result:
      seriesList.extend(result)

  return seriesList


def evaluateTokens(requestContext, tokens, replacements=None, pipedArg=None):
  if tokens.template:
    arglist = dict()
    if tokens.template.kwargs:
      arglist.update(dict([(kwarg.argname, evaluateScalarTokens(kwarg.args[0])) for kwarg in tokens.template.kwargs]))
    if tokens.template.args:
      arglist.update(dict([(str(i+1), evaluateScalarTokens(arg)) for i, arg in enumerate(tokens.template.args)]))
    if 'template' in requestContext:
      arglist.update(requestContext['template'])
    return evaluateTokens(requestContext, tokens.template, arglist)

  if tokens.expression:
    if tokens.expression.pipedCalls:
      # when the expression has piped calls, we pop the right-most call and pass the remaining
      # expression into it via pipedArg, to get the same result as a nested call
      rightMost = tokens.expression.pipedCalls.pop()
      return evaluateTokens(requestContext, rightMost, replacements, tokens)

    return evaluateTokens(requestContext, tokens.expression, replacements)

  if tokens.pathExpression:
    expression = tokens.pathExpression
    if replacements:
      for name in replacements:
        if expression == '$'+name:
          val = replacements[name]
          if not isinstance(val, six.string_types):
            return val
          elif re.match('^-?[\d.]+$', val):
            return float(val)
          else:
            return val
        else:
          expression = expression.replace('$'+name, str(replacements[name]))
    return fetchData(requestContext, expression)

  if tokens.call:
    if tokens.call.funcname == 'template':
      # if template propagates down here, it means the grammar didn't match the invocation
      # as tokens.template. this generally happens if you try to pass non-numeric/string args
      raise ValueError("invalid template() syntax, only string/numeric arguments are allowed")

    if tokens.call.funcname == 'seriesByTag':
      return fetchData(requestContext, tokens.call.raw)

    func = SeriesFunction(tokens.call.funcname)
    rawArgs = tokens.call.args or []
    if pipedArg is not None:
      rawArgs.insert(0, pipedArg)
    args = [evaluateTokens(requestContext, arg, replacements) for arg in rawArgs]
    requestContext['args'] = rawArgs
    kwargs = dict([(kwarg.argname, evaluateTokens(requestContext, kwarg.args[0], replacements))
                   for kwarg in tokens.call.kwargs])
    try:
      return func(requestContext, *args, **kwargs)
    except NormalizeEmptyResultError:
      return []

  return evaluateScalarTokens(tokens)


def evaluateScalarTokens(tokens):
  if tokens.number:
    if tokens.number.integer:
      return int(tokens.number.integer)
    if tokens.number.float:
      return float(tokens.number.float)
    if tokens.number.scientific:
      return float(tokens.number.scientific[0])

    raise ValueError("unknown numeric type in target evaluator")

  if tokens.string:
    return tokens.string[1:-1]

  if tokens.boolean:
    return tokens.boolean[0] == 'true'

  if tokens.none:
    return None

  raise ValueError("unknown token in target evaluator")


def extractPathExpressions(requestContext, targets):
  # Returns a list of unique pathExpressions found in the targets list

  pathExpressions = set()

  def extractPathExpression(requestContext, tokens, replacements=None):
    if tokens.template:
      arglist = dict()
      if tokens.template.kwargs:
        arglist.update(dict([(kwarg.argname, evaluateScalarTokens(kwarg.args[0])) for kwarg in tokens.template.kwargs]))
      if tokens.template.args:
        arglist.update(dict([(str(i+1), evaluateScalarTokens(arg)) for i, arg in enumerate(tokens.template.args)]))
      if 'template' in requestContext:
        arglist.update(requestContext['template'])
      extractPathExpression(requestContext, tokens.template, arglist)
    elif tokens.expression:
      extractPathExpression(requestContext, tokens.expression, replacements)
      if tokens.expression.pipedCalls:
        for token in tokens.expression.pipedCalls:
          extractPathExpression(requestContext, token, replacements)
    elif tokens.pathExpression:
      expression = tokens.pathExpression
      if replacements:
        for name in replacements:
          if expression != '$'+name:
            expression = expression.replace('$'+name, str(replacements[name]))
      pathExpressions.add(expression)
    elif tokens.call:
      # if we're prefetching seriesByTag, pass the entire call back as a path expression
      if tokens.call.funcname == 'seriesByTag':
        pathExpressions.add(tokens.call.raw)
      else:
        for a in tokens.call.args:
          extractPathExpression(requestContext, a, replacements)

  for target in targets:
    if not target:
      continue

    if isinstance(target, six.string_types):
      if not target.strip():
        continue
      target = grammar.parseString(target)
    extractPathExpression(requestContext, target)

  return list(pathExpressions)



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import os
from smtplib import SMTP
from socket import gethostname
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from six.moves.http_client import HTTPConnection
from six.moves.urllib.parse import urlsplit
from time import ctime, strftime
from traceback import format_exc
from graphite.user_util import getProfile
from graphite.logger import log
from graphite.account.models import MyGraph

from django.shortcuts import render_to_response
from django.http import HttpResponse
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist


def composer(request):
  profile = getProfile(request)
  context = {
    'queryString' : request.GET.urlencode().replace('+','%20'),
    'showTarget' : request.GET.get('showTarget',''),
    'user' : request.user,
    'profile' : profile,
    'showMyGraphs' : int( profile.user.username != 'default' ),
    'searchEnabled' : int( os.access(settings.INDEX_FILE, os.R_OK) ),
    'refreshInterval': settings.AUTO_REFRESH_INTERVAL,
    'debug' : settings.DEBUG,
    'jsdebug' : settings.DEBUG,
  }
  return render_to_response("composer.html",context)


def mygraph(request):
  profile = getProfile(request, allowDefault=False)

  if not profile:
    return HttpResponse( "You are not logged in!" )

  action = request.GET['action']
  graphName = request.GET['graphName']

  if not graphName:
    return HttpResponse("You must type in a graph name.")

  if action == 'save':
    url = request.GET['url']

    try:
      existingGraph = profile.mygraph_set.get(name=graphName)
      existingGraph.url = url
      existingGraph.save()

    except ObjectDoesNotExist:
      try:
        newGraph = MyGraph(profile=profile,name=graphName,url=url)
        newGraph.save()
      except Exception:
        log.exception("Failed to create new MyGraph in /composer/mygraph/, graphName=%s" % graphName)
        return HttpResponse("Failed to save graph %s" % graphName)

    return HttpResponse("SAVED")

  elif action == 'delete':
    try:
      existingGraph = profile.mygraph_set.get(name=graphName)
      existingGraph.delete()

    except ObjectDoesNotExist:
      return HttpResponse("No such graph '%s'" % graphName)

    return HttpResponse("DELETED")

  else:
    return HttpResponse("Invalid operation '%s'" % action)


def send_email(request):
  try:
    recipients = request.GET['to'].split(',')
    url = request.GET['url']
    proto, server, path, query, frag = urlsplit(url)
    if query: path += '?' + query
    conn = HTTPConnection(server)
    conn.request('GET',path)
    try: # Python 2.7+, use buffering of HTTP responses
      resp = conn.getresponse(buffering=True)
    except TypeError:  # Python 2.6 and older
      resp = conn.getresponse()
    assert resp.status == 200, "Failed HTTP response %s %s" % (resp.status, resp.reason)
    rawData = resp.read()
    conn.close()
    message = MIMEMultipart()
    message['Subject'] = "Graphite Image"
    message['To'] = ', '.join(recipients)
    message['From'] = 'composer@%s' % gethostname()
    text = MIMEText( "Image generated by the following graphite URL at %s\r\n\r\n%s" % (ctime(),url) )
    image = MIMEImage( rawData )
    image.add_header('Content-Disposition', 'attachment', filename="composer_" + strftime("%b%d_%I%M%p.png"))
    message.attach(text)
    message.attach(image)
    s = SMTP(settings.SMTP_SERVER)
    s.sendmail('composer@%s' % gethostname(),recipients,message.as_string())
    s.quit()
    return HttpResponse( "OK" )
  except Exception:
    return HttpResponse(format_exc())



from __future__ import absolute_import

from binascii import b2a_base64
import sys

from graphite.http_pool import http
from graphite.util import json

from graphite.tags.base import BaseTagDB


class HttpTagDB(BaseTagDB):
  """
  Stores tag information using an external http service that implements the graphite tags API.
  """
  def __init__(self, settings, *args, **kwargs):
    super(HttpTagDB, self).__init__(settings, *args, **kwargs)

    self.base_url = settings.TAGDB_HTTP_URL
    self.username = settings.TAGDB_HTTP_USER
    self.password = settings.TAGDB_HTTP_PASSWORD

  def request(self, method, url, fields, requestContext=None):
    headers = requestContext.get('forwardHeaders') if requestContext else {}
    if 'Authorization' not in headers and self.username and self.password:
      user_pw = '%s:%s' % (self.username, self.password)
      if sys.version_info[0] >= 3:
        user_pw_b64 = b2a_base64(user_pw.encode('utf-8')).decode('ascii')
      else:
        user_pw_b64 = user_pw.encode('base64')
      headers['Authorization'] = 'Basic ' + user_pw_b64

    req_fields = []
    for (field, value) in fields.items():
      if value is None:
        continue

      if isinstance(value, list) or isinstance(value, tuple):
        req_fields.extend([(field, v) for v in value if v is not None])
      else:
        req_fields.append((field, value))

    result = http.request(
      method,
      self.base_url + url,
      fields=req_fields,
      headers=headers,
      timeout=self.settings.FIND_TIMEOUT,
    )

    if result.status == 400:
      raise ValueError(json.loads(result.data.decode('utf-8')).get('error'))

    if result.status != 200:
      raise Exception('HTTP Error from remote tagdb: %s %s' % (result.status, result.data))

    return json.loads(result.data.decode('utf-8'))

  def find_series_cachekey(self, tags, requestContext=None):
    headers = [
      header + '=' + value
      for (header, value)
      in (requestContext.get('forwardHeaders', {}) if requestContext else {}).items()
    ]

    return 'TagDB.find_series:' + ':'.join(sorted(tags)) + ':' + ':'.join(sorted(headers))

  def _find_series(self, tags, requestContext=None):
    return self.request(
      'POST',
      '/tags/findSeries',
      {'expr': tags},
      requestContext=requestContext,
    )

  def get_series(self, path, requestContext=None):
    parsed = self.parse(path)

    seriesList = self.find_series(
      [('%s=%s' % (tag, parsed.tags[tag])) for tag in parsed.tags],
      requestContext=requestContext,
    )

    if parsed.path in seriesList:
      return parsed

  def list_tags(self, tagFilter=None, limit=None, requestContext=None):
    return self.request('GET', '/tags', {'filter': tagFilter, 'limit': limit}, requestContext)

  def get_tag(self, tag, valueFilter=None, limit=None, requestContext=None):
    return self.request('GET', '/tags/' + tag, {'filter': valueFilter, 'limit': limit}, requestContext)

  def list_values(self, tag, valueFilter=None, limit=None, requestContext=None):
    tagInfo = self.get_tag(tag, valueFilter=valueFilter, limit=limit, requestContext=requestContext)
    if not tagInfo:
      return []

    return tagInfo['values']

  def tag_series(self, series, requestContext=None):
    return self.request('POST', '/tags/tagSeries', {'path': series}, requestContext)

  def tag_multi_series(self, seriesList, requestContext=None):
    return self.request('POST', '/tags/tagMultiSeries', {'path': seriesList}, requestContext)

  def del_series(self, series, requestContext=None):
    return self.request('POST', '/tags/delSeries', {'path': series}, requestContext)

  def del_multi_series(self, seriesList, requestContext=None):
    return self.request('POST', '/tags/delSeries', {'path': seriesList}, requestContext)

  def auto_complete_tags(self, exprs, tagPrefix=None, limit=None, requestContext=None):
    """
    Return auto-complete suggestions for tags based on the matches for the specified expressions, optionally filtered by tag prefix
    """
    if not self.settings.TAGDB_HTTP_AUTOCOMPLETE:
      return super(HttpTagDB, self).auto_complete_tags(
        exprs, tagPrefix=tagPrefix, limit=limit, requestContext=requestContext)

    if limit is None:
      limit = self.settings.TAGDB_AUTOCOMPLETE_LIMIT

    fields = {
      'tagPrefix': tagPrefix or '',
      'limit': str(limit),
      'expr': exprs,
    }

    return self.request('POST', '/tags/autoComplete/tags', fields, requestContext)

  def auto_complete_values(self, exprs, tag, valuePrefix=None, limit=None, requestContext=None):
    """
    Return auto-complete suggestions for tags and values based on the matches for the specified expressions, optionally filtered by tag and/or value prefix
    """
    if not self.settings.TAGDB_HTTP_AUTOCOMPLETE:
      return super(HttpTagDB, self).auto_complete_values(
        exprs, tag, valuePrefix=valuePrefix, limit=limit, requestContext=requestContext)

    if limit is None:
      limit = self.settings.TAGDB_AUTOCOMPLETE_LIMIT

    fields = {
      'tag': tag or '',
      'valuePrefix': valuePrefix or '',
      'limit': str(limit),
      'expr': exprs,
    }

    return self.request('POST', '/tags/autoComplete/values', fields, requestContext)



INFINITY = float('inf')
NEGATIVE_INFINITY = -INFINITY


class IntervalSet:
  __slots__ = ('intervals', 'size')

  def __init__(self, intervals, disjoint=False):
    self.intervals = intervals

    if not disjoint:
      self.intervals = union_overlapping(self.intervals)

    self.size = sum(i.size for i in self.intervals)

  def __repr__(self):
    return repr(self.intervals)

  def __iter__(self):
    return iter(self.intervals)

  def __len__(self):
    return len(self.intervals)

  def __getitem__(self, i):
    return self.intervals[i]

  def __nonzero__(self):
    return self.size != 0

  def __sub__(self, other):
    return self.intersect( other.complement() )

  def complement(self):
    complementary = []
    cursor = NEGATIVE_INFINITY

    for interval in self.intervals:
      if cursor < interval.start:
        complementary.append( Interval(cursor, interval.start) )
        cursor = interval.end

    if cursor < INFINITY:
      complementary.append( Interval(cursor, INFINITY) )

    return IntervalSet(complementary, disjoint=True)

  def intersect(self, other): #XXX The last major bottleneck. Factorial-time hell.
    # Then again, this function is entirely unused...
    if (not self) or (not other):
      return IntervalSet([])

    #earliest = max(self.intervals[0].start, other.intervals[0].start)
    #latest = min(self.intervals[-1].end, other.intervals[-1].end)

    #mine = [i for i in self.intervals if i.start >= earliest and i.end <= latest]
    #theirs = [i for i in other.intervals if i.start >= earliest and i.end <= latest]

    intersections = [x for x in (i.intersect(j)
                                 for i in self.intervals
                                 for j in other.intervals)
                     if x]

    return IntervalSet(intersections, disjoint=True)

  def intersect_interval(self, interval):
    intersections = [x for x in (i.intersect(interval)
                                 for i in self.intervals)
                     if x]
    return IntervalSet(intersections, disjoint=True)

  def union(self, other):
    return IntervalSet( sorted(self.intervals + other.intervals) )


class Interval:
  __slots__ = ('start', 'end', 'tuple', 'size')

  def __init__(self, start, end):
    if end - start < 0:
      raise ValueError("Invalid interval start=%s end=%s" % (start, end))

    self.start = start
    self.end = end
    self.tuple = (start, end)
    self.size = self.end - self.start

  def __eq__(self, other):
    return self.tuple == other.tuple

  def __ne__(self, other):
    return self.tuple != other.tuple

  def __hash__(self):
    return hash( self.tuple )

  def __lt__(self, other):
    return self.start < self.start

  def __le__(self, other):
    return self.start <= self.start

  def __gt__(self, other):
    return self.start > self.start

  def __ge__(self, other):
    return self.start >= self.start

  def __cmp__(self, other):
    return (self.start > other.start) - (self.start < other.start)

  def __len__(self):
    raise TypeError("len() doesn't support infinite values, use the 'size' attribute instead")

  def __nonzero__(self):  # Python 2
    return self.size != 0

  def __bool__(self):  # Python 3
    return self.size != 0

  def __repr__(self):
    return '<Interval: %s>' % str(self.tuple)

  def intersect(self, other):
    start = max(self.start, other.start)
    end = min(self.end, other.end)

    if end > start:
      return Interval(start, end)

  def overlaps(self, other):
    earlier = self if self.start <= other.start else other
    later = self if earlier is other else other
    return earlier.end >= later.start

  def union(self, other):
    if not self.overlaps(other):
      raise TypeError("Union of disjoint intervals is not an interval")

    start = min(self.start, other.start)
    end = max(self.end, other.end)
    return Interval(start, end)


def union_overlapping(intervals):
  """Union any overlapping intervals in the given set."""
  disjoint_intervals = []

  for interval in intervals:
    if disjoint_intervals and disjoint_intervals[-1].overlaps(interval):
      disjoint_intervals[-1] = disjoint_intervals[-1].union(interval)
    else:
      disjoint_intervals.append(interval)

  return disjoint_intervals



from graphite.util import jsonResponse, HttpResponse, HttpError
from graphite.storage import STORE, extractForwardHeaders


def _requestContext(request, queryParams):
  return {
    'forwardHeaders': extractForwardHeaders(request),
    'localOnly': queryParams.get('local') == '1',
  }


@jsonResponse
def tagSeries(request, queryParams):
  if request.method != 'POST':
    return HttpResponse(status=405)

  path = queryParams.get('path')
  if not path:
    raise HttpError('no path specified', status=400)

  return STORE.tagdb.tag_series(path, requestContext=_requestContext(request, queryParams))


@jsonResponse
def tagMultiSeries(request, queryParams):
  if request.method != 'POST':
    return HttpResponse(status=405)

  paths = []
  # Normal format: ?path=name;tag1=value1;tag2=value2&path=name;tag1=value2;tag2=value2
  if len(queryParams.getlist('path')) > 0:
    paths = queryParams.getlist('path')
  # Rails/PHP/jQuery common practice format: ?path[]=...&path[]=...
  elif len(queryParams.getlist('path[]')) > 0:
    paths = queryParams.getlist('path[]')
  else:
    raise HttpError('no paths specified',status=400)

  return STORE.tagdb.tag_multi_series(paths, requestContext=_requestContext(request, queryParams))


@jsonResponse
def delSeries(request, queryParams):
  if request.method != 'POST':
    return HttpResponse(status=405)

  paths = []
  # Normal format: ?path=name;tag1=value1;tag2=value2&path=name;tag1=value2;tag2=value2
  if len(queryParams.getlist('path')) > 0:
    paths = queryParams.getlist('path')
  # Rails/PHP/jQuery common practice format: ?path[]=...&path[]=...
  elif len(queryParams.getlist('path[]')) > 0:
    paths = queryParams.getlist('path[]')
  else:
    raise HttpError('no path specified', status=400)

  return STORE.tagdb.del_multi_series(paths, requestContext=_requestContext(request, queryParams))


@jsonResponse
def findSeries(request, queryParams):
  if request.method not in ['GET', 'POST']:
    return HttpResponse(status=405)

  exprs = []
  # Normal format: ?expr=tag1=value1&expr=tag2=value2
  if len(queryParams.getlist('expr')) > 0:
    exprs = queryParams.getlist('expr')
  # Rails/PHP/jQuery common practice format: ?expr[]=tag1=value1&expr[]=tag2=value2
  elif len(queryParams.getlist('expr[]')) > 0:
    exprs = queryParams.getlist('expr[]')

  if not exprs:
    raise HttpError('no tag expressions specified', status=400)

  return STORE.tagdb.find_series(exprs, requestContext=_requestContext(request, queryParams))


@jsonResponse
def tagList(request, queryParams):
  if request.method != 'GET':
    return HttpResponse(status=405)

  return STORE.tagdb.list_tags(
    tagFilter=request.GET.get('filter'),
    limit=request.GET.get('limit'),
    requestContext=_requestContext(request, queryParams),
  )


@jsonResponse
def tagDetails(request, queryParams, tag):
  if request.method != 'GET':
    return HttpResponse(status=405)

  return STORE.tagdb.get_tag(
    tag,
    valueFilter=queryParams.get('filter'),
    limit=queryParams.get('limit'),
    requestContext=_requestContext(request, queryParams),
  )


@jsonResponse
def autoCompleteTags(request, queryParams):
  if request.method not in ['GET', 'POST']:
    return HttpResponse(status=405)

  exprs = []
  # Normal format: ?expr=tag1=value1&expr=tag2=value2
  if len(queryParams.getlist('expr')) > 0:
    exprs = queryParams.getlist('expr')
  # Rails/PHP/jQuery common practice format: ?expr[]=tag1=value1&expr[]=tag2=value2
  elif len(queryParams.getlist('expr[]')) > 0:
    exprs = queryParams.getlist('expr[]')

  return STORE.tagdb_auto_complete_tags(
    exprs,
    tagPrefix=queryParams.get('tagPrefix'),
    limit=queryParams.get('limit'),
    requestContext=_requestContext(request, queryParams)
  )


@jsonResponse
def autoCompleteValues(request, queryParams):
  if request.method not in ['GET', 'POST']:
    return HttpResponse(status=405)

  exprs = []
  # Normal format: ?expr=tag1=value1&expr=tag2=value2
  if len(queryParams.getlist('expr')) > 0:
    exprs = queryParams.getlist('expr')
  # Rails/PHP/jQuery common practice format: ?expr[]=tag1=value1&expr[]=tag2=value2
  elif len(queryParams.getlist('expr[]')) > 0:
    exprs = queryParams.getlist('expr[]')

  tag = queryParams.get('tag')
  if not tag:
    raise HttpError('no tag specified', status=400)

  return STORE.tagdb_auto_complete_values(
    exprs,
    tag,
    valuePrefix=queryParams.get('valuePrefix'),
    limit=queryParams.get('limit'),
    requestContext=_requestContext(request, queryParams)
  )



class ParamTypes(object):
  pass


for paramType in [
  'aggFunc',
  'boolean',
  'date',
  'float',
  'integer',
  'interval',
  'intOrInterval',
  'node',
  'nodeOrTag',
  'series',
  'seriesList',
  'seriesLists',
  'string',
  'tag',
]:
  setattr(ParamTypes, paramType, paramType)


class Param(object):
  __slots__ = ('name', 'type', 'required', 'default', 'multiple', 'options', 'suggestions')

  def __init__(self, name, paramtype, required=False, default=None, multiple=False, options=None,
               suggestions=None):
    self.name = name
    if not hasattr(ParamTypes, paramtype):
      raise Exception('Invalid type %s for parameter %s' % (paramtype, name))
    self.type = paramtype
    self.required = bool(required)
    self.default = default
    self.multiple = bool(multiple)
    self.options = options
    self.suggestions = suggestions

  def toJSON(self):
    jsonVal = {
      'name': self.name,
      'type': self.type,
    }
    if self.required:
      jsonVal['required'] = True
    if self.default is not None:
      jsonVal['default'] = self.default
    if self.multiple:
      jsonVal['multiple'] = True
    if self.options:
      jsonVal['options'] = self.options
    if self.suggestions:
      jsonVal['suggestions'] = self.suggestions
    return jsonVal
